#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Union, Callable, Tuple

import numpy as np
import pyopencl as cl
from os.path import dirname, join

import warnings
from pyopencl import CompilerWarning

from numpy import array, uint32, float32, int32, asarray, zeros, ones, unique, atleast_2d, squeeze, ndarray, \
    concatenate, empty, linspace, diff, trapz
from .ldmodel import LDModel

from pytransit.models.numba.rrmodel import create_z_grid

from .transitmodel import TransitModel
from .numba.ldmodels import *

warnings.filterwarnings('ignore', category=CompilerWarning)

__all__ = ['SwiftModelCL']

class SwiftModelCL(TransitModel):
    ldmodels = {'uniform': (ld_uniform, ldi_uniform),
                'linear': (ld_linear, ldi_linear),
                'quadratic': (ld_quadratic, ldi_quadratic),
                'quadratic-tri': (ld_quadratic_tri, ldi_quadratic_tri),
                'nonlinear': ld_nonlinear,
                'general': ld_general,
                'square_root': ld_square_root,
                'logarithmic': ld_logarithmic,
                'exponential': ld_exponential,
                'power-2': ld_power_2,
                'power-2-pm': ld_power_2_pm}

    def __init__(self, ldmodel: Union[str, Callable, Tuple[Callable, Callable]] = 'quadratic',
                 interpolate: bool = False, klims: tuple = (0.005, 0.5), nk: int = 256,
                 nzin: int = 20, nzlimb: int = 20, zcut=0.7, ng: int = 50, parallel: bool = False,
                 small_planet_limit: float = 0.05, cl_ctx=None, cl_queue=None) -> None:
        super().__init__()

        self.ctx = cl_ctx or cl.create_some_context()
        self.queue = cl_queue or cl.CommandQueue(self.ctx)

        self.splimit = small_planet_limit

        # Set up the limb darkening model
        # --------------------------------
        if isinstance(ldmodel, str):
            try:
                if isinstance(self.ldmodels[ldmodel], tuple):
                    self.ldmodel = self.ldmodels[ldmodel][0]
                    self.ldmmean = self.ldmodels[ldmodel][1]
                else:
                    self.ldmodel = self.ldmodels[ldmodel]
                    self.ldmmean = None
            except KeyError:
                print(
                    f"Unknown limb darkening model: {ldmodel}. Choose from [{', '.join(self.ldmodels.keys())}] or supply a callable function.")
                raise
        elif isinstance(ldmodel, LDModel):
            self.ldmodel = ldmodel
            self.ldmmean = ldmodel._integrate
        elif callable(ldmodel):
            self.ldmodel = ldmodel
            self.ldmmean = None
        elif isinstance(ldmodel, tuple) and callable(ldmodel[0]) and callable(ldmodel[1]):
            self.ldmodel = ldmodel[0]
            self.ldmmean = ldmodel[1]
        else:
            raise NotImplementedError

        self._ldmu = linspace(1, 0, 200)
        self._ldz = sqrt(1 - self._ldmu ** 2)

        # Set the basic variable
        # ----------------------
        self.klims = klims
        self.nk = nk
        self.ng = ng
        self.nzin = nzin
        self.nzlimb = nzlimb
        self.zcut = zcut

        self.npv = None
        self.nptb  = 0
        self.npb   = 0
        self.f     = None
        self.pv = array([])

        self.time  = None
        self.lcids = None
        self.pbids = None
        self.nsamples = None
        self.exptimes = None

        self.ze = None
        self.gs = None
        self.dg = None

        # Declare the buffers for the swift model arrays
        self._b_ze = None
        self._b_gs = None
        self._b_weights = None
        self._b_istar = None
        self._b_ldp = None
        self._b_ldw = None
        self._b_ks = None

        # Declare the buffers for the ld coefficients, time, and flux arrays. These will
        # be initialised when the model is first evaluated, and reinitialised if the
        # array sizes change.
        #
        self._b_time = None    # Time buffer
        self._b_f = None       # Flux buffer
        self._b_p = None       # Parameter vector buffer

        self._time_id = None   # Time array ID

        self.prg = cl.Program(self.ctx, open(join(dirname(__file__),'opencl','swiftmodel.cl'),'r').read()).build()

        self.init_siwft_arrays(self.zcut, self.ng, self.nzin, self.nzlimb)

    def init_siwft_arrays(self, zcut: float = 0.7, ng: int = 50, nzin: int = 30, nzlimb: int = 30):
        mf = cl.mem_flags

        self.ze, self.zm = create_z_grid(zcut, nzin, nzlimb)
        self.mu = sqrt(1-self.zm**2).astype('float32')
        self.ze = self.ze.astype('float32')
        self.nz = int32(self.ze.size)
        self.ng = int32(ng)

        self.gs = linspace(0, 0.9999, ng).astype('float32')
        self.dg = float32(diff(self.gs)[0])

        if self._b_ze is not None:
            self._b_ze.release()
            self._b_gs.releare()

        self._b_ze = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.ze)
        self._b_gs = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.gs)

    def set_data(self, time, lcids=None, pbids=None, nsamples=None, exptimes=None):
        mf = cl.mem_flags

        if self._b_time is not None:
            self._b_time.release()
            self._b_lcids.release()
            self._b_pbids.release()
            self._b_nsamples.release()
            self._b_etimes.release()

        self.nlc = uint32(1 if lcids is None else unique(lcids).size)
        self.npb = uint32(1 if pbids is None else unique(pbids).size)
        self.nptb = time.size

        self.time = asarray(time, dtype='float32')
        self.lcids = zeros(time.size, 'uint32') if lcids is None else asarray(lcids, dtype='uint32')
        self.pbids = zeros(self.nlc, 'uint32') if pbids is None else asarray(pbids, dtype='uint32')
        self.nsamples = ones(self.nlc, 'uint32') if nsamples is None else asarray(nsamples, dtype='uint32')
        self.exptimes = ones(self.nlc, 'float32') if exptimes is None else asarray(exptimes, dtype='float32')

        self._b_time = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.time)
        self._b_lcids = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.lcids)
        self._b_pbids = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.pbids)
        self._b_nsamples = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.nsamples)
        self._b_etimes = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.exptimes)


    def evaluate(self, k: Union[float, ndarray], ldc: ndarray, t0: Union[float, ndarray], p: Union[float, ndarray],
                 a: Union[float, ndarray], i: Union[float, ndarray], e: Optional[Union[float, ndarray]] = None,
                 w: Optional[Union[float, ndarray]] = None, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a set of scalar or vector parameters.

        Parameters
        ----------
        k
            Radius ratio(s) either as a single float, 1D vector, or 2D array.
        ldc
            Limb darkening coefficients as a 1D or 2D array.
        t0
            Transit center(s) as a float or a 1D vector.
        p
            Orbital period(s) as a float or a 1D vector.
        a
            Orbital semi-major axis (axes) divided by the stellar radius as a float or a 1D vector.
        i
            Orbital inclination(s) as a float or a 1D vector.
        e : optional
            Orbital eccentricity as a float or a 1D vector.
        w : optional
            Argument of periastron as a float or a 1D vector.

        Notes
        -----
        The model can be evaluated either for one set of parameters or for many sets of parameters simultaneously. In
        the first case, the orbital parameters should all be given as floats. In the second case, the orbital parameters
        should be given as a 1D array-like.

        Returns
        -------
        ndarray
            Modelled flux either as a 1D or 2D ndarray.
        """
        npv = 1 if isinstance(t0, float) else len(t0)
        k = asarray(k)

        if k.size == 1:
            nk = 1
        elif npv == 1:
            nk = k.size
        else:
            nk = k.shape[1]

        if e is None:
            e, w = 0.0, 0.0

        pvp = empty((npv, nk + 6), dtype=float32)
        pvp[:, :nk] = k
        pvp[:, nk] = t0
        pvp[:, nk + 1] = p
        pvp[:, nk + 2] = a
        pvp[:, nk + 3] = i
        pvp[:, nk + 4] = e
        pvp[:, nk + 5] = w

        return self.evaluate_pv(pvp, ldc, copy)

    def evaluate_ps(self, k, ldc, t0, p, a, i, e=0., w=0., copy=True) -> ndarray:
        """Evaluate the transit model for a set of scalar parameters.

           Parameters
           ----------
           k : array-like
               Radius ratio(s) either as a single float or an 1D array.
           ldc
             Limb darkening coefficients as a 1D or 2D array.
           t0 : float
               Transit center as a float.
           p : float
               Orbital period as a float.
           a : float
               Orbital semi-major axis divided by the stellar radius as a float.
           i : float
               Orbital inclination(s) as a float.
           e : float, optional
               Orbital eccentricity as a float.
           w : float, optional
               Argument of periastron as a float.

           Notes
           -----
           This version of the `evaluate` method is optimized for calculating a single transit model (such as when using a
           local optimizer). If you want to evaluate the model for a large number of parameters simultaneously, use either
           `evaluate` or `evaluate_pv`.

           Returns
           -------
           ndarray
               Modelled flux as a 1D ndarray.
           """
        if isinstance(k, float):
            pv = array([[k, t0, p, a, i, e, w]], float32)
        else:
            pv = concatenate([k, [t0, p, a, i, e, w]]).astype(float32)
        return self.evaluate_pv(pv, ldc, copy)

    def evaluate_pv(self, pvp: ndarray, ldc: ndarray, copy: bool = True) -> ndarray:
        """Evaluate the transit model for a 2D parameter array.

           Parameters
           ----------
           pvp
               Parameter array with a shape `(npv, npar)` where `npv` is the number of parameter vectors, and each row
               contains a set of parameters `[k, t0, p, a, i, e, w]`. The radius ratios can also be given per passband,
               in which case the row should be structured as `[k_0, k_1, k_2, ..., k_npb, t0, p, a, i, e, w]`.

           Notes
           -----
           This version of the `evaluate` method is optimized for calculating several models in parallel, such as when
           using *emcee* for MCMC sampling.

           Returns
           -------
           ndarray
               Modelled flux either as a 1D or 2D ndarray.
           """
        pvp = atleast_2d(pvp)
        ldc = asarray(ldc)
        nk = pvp.shape[1] - 6

        # Release and reinitialise the GPU buffers if the parameter vector size changes
        if self.npv != pvp.shape[0]:
            self.npv = uint32(pvp.shape[0])
            self.spv = uint32(pvp.shape[1])

            if self._b_f is not None:
                self._b_f.release()
                self._b_p.release()

            if self._b_weights is not None:
                self._b_weights.release()
                self._b_ldp.release()
                self._b_ldw.release()
                self._b_istar.release()
                self._b_ks.release()

            self.pv = zeros(pvp.shape, float32)
            self.f = zeros((self.npv, self.nptb), float32)

            mf = cl.mem_flags
            self._b_f = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.time.nbytes * self.npv)
            self._b_p = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=self.pv)
            self._b_weights = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.npv * self.ng * self.nz * float32().nbytes)
            self._b_ldp = cl.Buffer(self.ctx, mf.READ_ONLY, self.npv*self.npb*self.ng*float32().nbytes)
            self._b_ldw = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.npv*self.npb*self.ng*float32().nbytes)
            self._b_istar = cl.Buffer(self.ctx, mf.WRITE_ONLY, self.npv*self.npb*float32().nbytes)
            self._b_ks = cl.Buffer(self.ctx, mf.READ_ONLY, self.npv*float32().nbytes)

        if isinstance(self.ldmodel, LDModel):
            ldp, istar = self.ldmodel(self.mu, ldc)
        else:
            ldp = evaluate_ld(self.ldmodel, self.mu, ldc)

            if self.ldmmean is not None:
                istar = evaluate_ldi(self.ldmmean, ldc)
            else:
                istar = zeros((self.npv, self.npb))
                ldpi = evaluate_ld(self.ldmodel, self._ldmu, ldc)
                for ipv in range(self.npv):
                    for ipb in range(self.npb):
                        istar[ipv, ipb] = 2 * pi * trapz(self._ldz * ldpi[ipv,ipb], self._ldz)

        # Copy the limb darkening profiles and their integrals to the GPU
        cl.enqueue_copy(self.queue, self._b_ldp, ldp.astype('float32'))
        cl.enqueue_copy(self.queue, self._b_istar, istar.astype('float32'))
        cl.enqueue_copy(self.queue, self._b_ks, pvp[:, :nk].mean(1).astype('float32'))

        # Copy the parameter vector to the GPU
        self.pv[:] = pvp
        cl.enqueue_copy(self.queue, self._b_p, self.pv)

        self.prg.calculate_weights(self.queue, (self.npv, self.ng, self.nz), None, self._b_ks, self._b_ze, self._b_gs, self._b_weights)
        self.prg.calculate_ldw(self.queue, (self.npv, self.npb, self.ng), None, self.nz, self._b_ldp, self._b_weights, self._b_ldw)

        self.prg.swift_pop(self.queue, (self.npv, self.nptb), None, self._b_time, self._b_istar, self._b_ldw, self.ng,
                           self.dg, self._b_lcids, self._b_pbids, self._b_p, self._b_nsamples, self._b_etimes,
                           self.spv, self.nlc, self.npb, self._b_f)

        if copy:
            cl.enqueue_copy(self.queue, self.f, self._b_f)
            return squeeze(self.f)
        else:
            return None


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union

from numpy import array, diag, zeros, log10
from numpy.random import normal, multivariate_normal
from uncertainties import UFloat


def md_rs_from_rho(rho: Union[UFloat, float], nsamples: int = 100):
    """M dwarf stellar radius from stellar density (Hartman et al. 2015)

    Calculates the stellar radius from a stellar density estimate using the
    empirical M dwarf density-radius relation presented by Hartman et al. (2015).

    Parameters
    ----------
    rho: float or UFloat
        Stellar density in g/cm^3
    nsamples
        Number of samples to compute

    Returns
    -------
    ndarray
        Stellar radius in Solar radii

    """

    arm = array([0.054, -0.2989, 0.0815, -0.18280])
    are = array([0.048, 0.0083, 0.0020, 0.00075])
    A = array([0.131674, -0.605624, 0.739127, -0.263765,
               -0.494952, 0.614479, 0.437886, -0.430922,
               -0.765425, -0.291096, 0.099548, 0.565224,
               0.389626, 0.413398, 0.502032, 0.652117]).reshape([4, 4])

    ars = multivariate_normal(arm, diag(are ** 2), size=nsamples)

    if isinstance(rho, UFloat):
        rhos = normal(rho.n, rho.s, nsamples)
        rs = zeros(nsamples)
        for i in range(nsamples):
            lrho = log10(rhos[i])
            lra = array([1, lrho, lrho ** 2, lrho ** 3])
            rs[i] = 10 ** (ars[i] @ A @ lra)
        return rs
    elif isinstance(rho, float):
        lrho = log10(rho)
        lra = array([1, lrho, lrho ** 2, lrho ** 3])
        return 10 ** (ars @ A @ lra)
    else:
        raise NotImplementedError



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Union, List, Tuple, Optional

import numpy as np
from matplotlib.pyplot import subplots, setp
from numba import njit
from numpy import inf, squeeze, atleast_2d, sqrt, zeros_like, \
    ones, arange, argsort, arctan2

from .tesslpf import downsample_time
from .. import BaseLPF, TransitModel
from ..orbits import as_from_rhop, i_from_ba, fold, i_from_baew, d_from_pkaiews
from ..param import UniformPrior as UP, NormalPrior as NP, GParameter

np.seterr(invalid='ignore')


@njit(fastmath=True, cache=False)
def map_ldc(ldc):
    ldc = atleast_2d(ldc)
    uv = zeros_like(ldc)
    a, b = sqrt(ldc[:, 0::2]), 2. * ldc[:, 1::2]
    uv[:, 0::2] = a * b
    uv[:, 1::2] = a * (1. - b)
    return uv


class MPTLPF(BaseLPF):
    """Transit LPF for multiplanet systems.
    """
    def __init__(self, name: str, nplanets: int, passbands: List[str],
                 times: List = None, fluxes: List = None, errors: List = None,
                 pbids: List = None, covariates: List = None, wnids: List = None,
                 tm: TransitModel = None,
                 nsamples: Union[List[int], int] = 1,
                 exptimes: Union[List[float], float] = 0.,
                 init_data=True, result_dir: Path = None,
                 tref: float = 0.0, lnlikelihood: str = 'wn'):

        self.nplanets = nplanets
        super().__init__(name, passbands, times, fluxes, errors, pbids, covariates, wnids, tm,
                         nsamples, exptimes, init_data, result_dir, tref, lnlikelihood)

    def _init_p_orbit(self):
        porbit = [GParameter('rho', 'stellar_density', 'g/cm^3', UP(0.1, 25.0), (0, inf))]
        self.ps.add_global_block('orbit', porbit)

    def _init_p_planet(self):
        ps = self.ps
        pp = []
        for i in range(1, self.nplanets + 1):
            pp.extend([
                GParameter(f'tc_{i}', f'zero_epoch_{i}', '', NP(0.0, 0.1), (-inf, inf)),
                GParameter(f'p_{i}', f'period_{i}', 'd', NP(1.0, 1e-5), (0, inf)),
                GParameter(f'b_{i}', f'impact_parameter_{i}', 'R_s', UP(0.0, 1.0), (0, inf)),
                GParameter(f'k2_{i}', f'area_ratio_{i}', 'A_s', UP(0.02 ** 2, 0.1 ** 2), (0, inf)),
                GParameter(f'secw_{i}', f'sqrt(e) cos(w) {i}', '', UP(-1.0, 1.0), (-1, 1)),
                GParameter(f'sesw_{i}', f'sqrt(e) sin(w) {i}', '', UP(-1.0, 1.0), (-1, 1)),
            ])
        ps.add_global_block('planets', pp)
        self._start_pl = ps.blocks[-1].start
        self._sl_pl = ps.blocks[-1].slice

    def transit_model(self, pv, copy=True, planets=None):
        pv = atleast_2d(pv)
        flux = ones([pv.shape[0], self.timea.size])
        ldc = map_ldc(pv[:, self._sl_ld])
        planets = planets if planets is not None else arange(self.nplanets)
        for i in planets:
            pp = pv[:, 1 + 6 * i: 7 + 6 * i]
            t0 = pp[:, 0]
            p = pp[:, 1]
            k = sqrt(pp[:, 3])
            aos = as_from_rhop(pv[:, 0], p)
            inc = i_from_ba(pp[:, 2], aos)
            ecc = pp[:, 4] ** 2 + pp[:, 5] ** 2
            w = arctan2(pp[:, 5], pp[:, 4])
            flux += self.tm.evaluate(k, ldc, t0, p, aos, inc, ecc, w, copy) - 1.
        return squeeze(flux)

    def create_pv_population(self, npop=50):
        return self.ps.sample_from_prior(npop)

    def posterior_samples(self, burn: int = 0, thin: int = 1, derived_parameters: bool = True):
        df = super(MPTLPF, self).posterior_samples(burn=burn, thin=thin, derived_parameters=False)
        if derived_parameters:
            for i in range(1, self.nplanets + 1):
                p = df[f'p_{i}'].values
                k = df[f'k_{i}'] = sqrt(df[f'k2_{i}'])
                e = df[f'e_{i}'] = df[f'secw_{i}'] ** 2 + df[f'sesw_{i}'] ** 2
                w = df[f'w_{i}'] = arctan2(df[f'sesw_{i}'], df[f'secw_{i}'])
                a = df[f'a_{i}'] = as_from_rhop(df.rho.values, p)
                inc = df[f'inc_{i}'] = i_from_baew(df[f'b_{i}'].values, a, e.values, w.values)
                df[f't14_{i}'] = d_from_pkaiews(p, k.values, a, inc, e.values, w.values, 1, kind=14)
                df[f't23_{i}'] = d_from_pkaiews(p, k.values, a, inc, e.values, w.values, 1, kind=23)
        return df

    def plot_folded_planets(self, bin_width: float = 10.0, window_width: float = 3.0, figsize: Optional[Tuple] = None):
        fig, axs = subplots(1, self.nplanets, figsize=figsize, sharey='all')
        pv = self.de.minimum_location.copy()
        for ipl in range(self.nplanets):
            planets = set(arange(self.nplanets))
            planets.remove(ipl)
            t0 = pv[1 + 6 * ipl]
            p = pv[2 + 6 * ipl]
            mflux = self.transit_model(pv, planets=[ipl])
            rflux = self.transit_model(pv, planets=planets)
            oflux = self.ofluxa / rflux
            phase = fold(self.timea, p, t0)
            m = abs(phase) < 0.5 * window_width / 24
            sids = argsort(phase[m])
            phase, mflux, oflux = phase[m][sids], mflux[m][sids], oflux[m][sids]
            bp, bf, be = downsample_time(phase, oflux, bin_width / 24 / 60)
            axs[ipl].errorbar(bp, bf, be, fmt='ok')
            axs[ipl].plot(phase, oflux, '.', alpha=0.25)
            axs[ipl].plot(phase, mflux, 'k')
        setp(axs, xlim=(-0.5 * window_width / 24, 0.5 * window_width / 24), ylim=(0.996, 1.004), xlabel='Time - T$_0$ [d]')
        setp(axs[0], ylabel='Normalised flux')
        fig.tight_layout()
        return fig



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union, List, Iterable

from numpy import inf, repeat, pi, atleast_2d, arctan2, sqrt, squeeze, ndarray, zeros
from ..models.numba.phasecurves import doppler_boosting, ellipsoidal_variation, emission, lambert_phase_function

from .. import QuadraticModel, EclipseModel
from ..orbits import as_from_rhop, i_from_ba
from ..param import GParameter, PParameter, NormalPrior as NP, UniformPrior as UP
from .lpf import BaseLPF, map_ldc


class PhaseCurveLPF(BaseLPF):
    def _init_data(self, times: Union[List, ndarray], fluxes: Union[List, ndarray], pbids: Union[List, ndarray] = None,
                   covariates: Union[List, ndarray] = None, errors: Union[List, ndarray] = None,
                   wnids: Union[List, ndarray] = None,
                   nsamples: Union[int, ndarray, Iterable] = 1, exptimes: Union[float, ndarray, Iterable] = 0.):

        super()._init_data(times=times, fluxes=fluxes, pbids=pbids, covariates=covariates,
                           errors=errors, wnids=wnids, nsamples=nsamples, exptimes=exptimes)

        pbis = self.pbids[self.lcids]
        self.pbmasks = []
        for i in range(self.npb):
            self.pbmasks.append(pbis == i)

    def _post_initialisation(self):
        self.tm = QuadraticModel(interpolate=False)
        self.em = EclipseModel()
        self.tm.set_data(self.timea - self._tref, self.lcids, self.pbids, self.nsamples, self.exptimes)
        self.em.set_data(self.timea - self._tref, self.lcids, self.pbids, self.nsamples, self.exptimes)

    def _init_p_orbit(self):
        porbit = [
            GParameter('tc', 'zero epoch', 'd', NP(0.0, 0.1), (-inf, inf)),
            GParameter('p', 'p', 'd', NP(1.0, 1e-5), (0, inf)),
            GParameter('rho', 'stellar density', 'g/cm^3', UP(0.1, 25.0), (0, inf)),
            GParameter('b', 'impact parameter', 'R_s', UP(0.0, 1.0), (0, 1)),
            GParameter('secw', 'sqrt(e) cos(w)', '', UP(-0.5, 0.5), (-1, 1)),
            GParameter('sesw', 'sqrt(e) sin(w)', '', UP(-0.5, 0.5), (-1, 1))]
        self.ps.add_global_block('orbit', porbit)

    def _init_p_planet(self):
        """Planet parameter initialisation.
        """
        pk2 = [PParameter('k2', 'area ratio', 'A_s', UP(0.01 ** 2, 0.2 ** 2), (0, inf))]
        self.ps.add_global_block('k2', pk2)
        self._pid_k2 = repeat(self.ps.blocks[-1].start, 1)
        self._start_k2 = self.ps.blocks[-1].start
        self._sl_k2 = self.ps.blocks[-1].slice
        self._ix_k2 = self._start_k2

        pph = [GParameter(f'oev', f'Ellipsoidal variation offset', '', NP(0, 1e-6), (-inf, inf))]
        for pb in self.passbands:
            pph.extend([GParameter(f'aev_{pb}', f'Ellipsoidal variation amplitude in {pb}', '', UP(0, 1), (0, inf)),
                        GParameter(f'adb_{pb}', f'Doppler boosting amplitude in {pb}', '', UP(0, 1), (0, inf)),
                        GParameter(f'ted_{pb}', f'Day-side flux ratio in {pb}', '', UP(0.0, 0.2), (-inf, inf)),
                        GParameter(f'ten_{pb}', f'Night-side flux ratio in {pb}', '', UP(0.0, 0.1), (-inf, inf)),
                        GParameter(f'teo_{pb}', 'Thermal emission offset', 'rad', UP(-pi, pi), (-inf, inf)),
                        GParameter(f'ag_{pb}', f'Geometric albedo in {pb}', '', UP(0, 1), (0, 1))])
        self.ps.add_global_block('phase', pph)
        self._pid_fr = repeat(self.ps.blocks[-1].start, 1)
        self._start_fr = self.ps.blocks[-1].start
        self._sl_fr = self.ps.blocks[-1].slice

    def map_pv(self, pv):
        pv = atleast_2d(pv)
        t0 = pv[:, 0]
        p = pv[:, 1]
        a = as_from_rhop(pv[:, 2], p)
        inc = i_from_ba(pv[:, 3], a)
        ecc = pv[:, 4] ** 2 + pv[:, 5] ** 2
        omega = arctan2(pv[:, 5], pv[:, 4])
        area_ratio = pv[:, 6]
        k = sqrt(pv[:, 6:7])
        ldc = map_ldc(pv[:, self._sl_ld])
        sle = self._sl_fr
        oev = pv[:, sle][:, 0]
        aev = pv[:, sle][:, 1::6]
        adb = pv[:, sle][:, 2::6]
        dte = pv[:, sle][:, 3::6]
        nte = pv[:, sle][:, 4::6]
        ote = pv[:, sle][:, 5::6]
        ag  = pv[:, sle][:, 6::6]
        return t0, p, a, inc, ecc, omega, area_ratio, k, ldc, oev, aev, adb, dte, nte, ote, ag

    def doppler_boosting(self, pv):
        t0, p, a, inc, ecc, omega, area_ratio, k, ldc, oev, aev, adb, dte, nte, ote, ab = self.map_pv(pv)
        npv, npt = inc.size, self.ofluxa.size
        fdb = zeros((npv, npt))
        for ipb, pbm in enumerate(self.pbmasks):
            fdb[:, pbm] = doppler_boosting(adb[:, ipb], t0, p, True, self.timea[pbm])
        return squeeze(fdb)

    def ellipsoidal_variations(self, pv):
        t0, p, a, inc, ecc, omega, area_ratio, k, ldc, oev, aev, adb, dte, nte, ote, ab = self.map_pv(pv)
        npv, npt = inc.size, self.ofluxa.size
        fev = zeros((npv, npt))
        for ipb, pbm in enumerate(self.pbmasks):
            fev[:, pbm] = ellipsoidal_variation(aev[:, ipb], t0, p, oev, True, self.timea[pbm])
        return squeeze(fev)

    def thermal_flux(self, pv):
        t0, p, a, inc, ecc, omega, area_ratio, k, ldc, oev, aev, adb, dte, nte, ote, ab = self.map_pv(pv)
        npv, npt = inc.size, self.ofluxa.size
        fec = self.em.evaluate(k, t0, p, a, inc, ecc, omega, multiplicative=True)
        ft = zeros((npv, npt))
        for ipb, pbm in enumerate(self.pbmasks):
            ft[:, pbm] = emission(area_ratio, nte[:, ipb], dte[:, ipb], ote[:, ipb], t0, p, True, self.timea[pbm])
        return fec * squeeze(ft)

    def reflected_flux(self, pv):
        t0, p, a, inc, ecc, omega, area_ratio, k, ldc, oev, aev, adb, dte, nte, ote, ab = self.map_pv(pv)
        npv, npt = inc.size, self.ofluxa.size
        fec = self.em.evaluate(k, t0, p, a, inc, ecc, omega, multiplicative=True)
        fr = zeros((npv, npt))
        for ipb, pbm in enumerate(self.pbmasks):
            fr[:, pbm] = lambert_phase_function(a, area_ratio, ab[:, ipb], t0, p, True, self.timea[pbm])
        return fec * squeeze(fr)

    def transit_model(self, pv, copy=True):
        t0, p, a, inc, ecc, omega, area_ratio, k, ldc, oev, aev, adb, dte, nte, ote, ab = self.map_pv(pv)
        npv, npt = inc.size, self.ofluxa.size

        fec = self.em.evaluate(k, t0 - self._tref, p, a, inc, ecc, omega, multiplicative=True)
        ft = zeros((npv, npt))
        fr = zeros((npv, npt))
        for ipb, pbm in enumerate(self.pbmasks):
            ft[:, pbm] = emission(area_ratio, nte[:, ipb], dte[:, ipb], ote[:, ipb], t0, p, True, self.timea[pbm])
            fr[:, pbm] = lambert_phase_function(a, area_ratio, ab[:, ipb], t0, p, True, self.timea[pbm])
        fplanet = fec * squeeze((ft + fr))

        ftr = self.tm.evaluate(k, ldc, t0 - self._tref, p, a, inc, ecc, omega)
        fev = zeros((npv, npt))
        fdb = zeros((npv, npt))
        for ipb, pbm in enumerate(self.pbmasks):
            fev[:, pbm] = ellipsoidal_variation(aev[:, ipb], t0, p, oev, True, self.timea[pbm])
            fdb[:, pbm] = doppler_boosting(adb[:, ipb], t0, p, True, self.timea[pbm])
        fstar = squeeze(ftr + fev + fdb)
        return fplanet + fstar

    def flux_model(self, pv):
        baseline = self.baseline(pv)
        model_flux = self.transit_model(pv)
        return squeeze(model_flux * atleast_2d(baseline))



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import seaborn as sb
from matplotlib.pyplot import subplots, setp
from numpy import pi, sign, cos, sqrt, sin, array, arccos, inf, round, int, s_, percentile, concatenate, median, mean, \
    arange, poly1d, polyfit

from numba import njit, prange
from .lpf import BaseLPF
from ..models.transitmodel import TransitModel
from ..param.parameter import ParameterSet, PParameter, GParameter
from ..param.parameter import UniformPrior as U, NormalPrior as N, GammaPrior as GM
from ..orbits.orbits_py import as_from_rhop

with_seaborn = True

@njit("f8[:](f8[:], f8, f8, f8, f8[:], i8[:])", cache=False, parallel=False)
def z_circular_ttv(t, p, a, i, tc, tcid):
    cosph = cos(2*pi * (t - tc[tcid]) / p)
    z = sign(cosph) * a * sqrt(1.0 - cosph * cosph * sin(i) ** 2)
    return z


def plot_estimates(x, p, ax, bwidth=0.8):
    ax.bar(x, p[4, :] - p[3, :], bwidth, p[3, :], alpha=0.25, fc='b')
    ax.bar(x, p[2, :] - p[1, :], bwidth, p[1, :], alpha=0.25, fc='b')
    [ax.plot((xx - 0.47 * bwidth, xx + 0.47 * bwidth), (pp[[0, 0]]), 'k') for xx, pp in zip(x, p.T)]

class TTVLPF(BaseLPF):
    """Log posterior function for TTV estimation.

    A log posterior function for TTV estimation. Each light curve represents a single transit, and
    is given a separate free transit centre parameter. The average orbital period and (one) transit
    zero epoch are assumed as known.

    Notes: The number of parameters can grow large with Kepler short-period planets.

    """
    def __init__(self, target: str, zero_epoch: float, period: float, tc_sigma: float, passbands: list,
                 times: list = None, fluxes: list = None, pbids: list = None, tm: TransitModel = None,
                 nsamples: int = 1, exptime: float = 0.020433598):
        self.zero_epoch = zero_epoch
        self.period = period
        self.tc_sigma = tc_sigma
        super().__init__(target, passbands, times, fluxes, pbids, tm, nsamples, exptime)

    def _init_p_orbit(self):
        """Orbit parameter initialisation for a TTV model.
        """
        porbit = [GParameter('rho', 'stellar_density', 'g/cm^3', U(0.1, 25.0), (0, inf)),
                  GParameter('b', 'impact_parameter', 'R_s', U(0.0, 1.0), (0, 1))]

        s = self.tc_sigma
        self.tnumber = round((array([t.mean() for t in self.times]) - self.zero_epoch) / self.period).astype(int)
        tcs = self.period * self.tnumber + self.zero_epoch
        for tc, tn in zip(tcs, self.tnumber):
            porbit.append(GParameter(f'tc_{tn:d}', f'transit_centre_{tn:d}', 'd', N(tc, s), (-inf, inf)))
        self.ps.add_global_block('orbit', porbit)
        self._start_tc = 2
        self._sl_tc = s_[self._start_tc:self._start_tc + self.nlc]

    def optimize_times(self, window):
        times, fluxes, pbids = [], [], []
        tcp = self.ps[self._sl_tc]
        for i in range(self.nlc):
            tc = tcp[i].prior.mean
            mask = abs(self.times[i] - tc) < 0.5*window/24.
            times.append(self.times[i][mask])
            fluxes.append(self.fluxes[i][mask])
        self._init_data(times, fluxes, self.pbids)

    def _compute_z(self, pv):
        a = as_from_rhop(pv[0], self.period)
        if a < 1.:
            return None
        else:
            i = arccos(pv[1] / a)
            tc = pv[self._sl_tc]
            return z_circular_ttv(self.timea, self.period, a, i, tc, self.lcida)

    def plot_light_curve(self, ax=None, figsize=None, time=False):
        fig, ax = (None, ax) if ax is not None else subplots(figsize=figsize)
        time = self.timea_orig if time else arange(self.timea_orig.size)
        ax.plot(time, concatenate(self.fluxes))
        ax.plot(time, concatenate(self.flux_model(self.de.minimum_location)))
        fig.tight_layout()
        return ax

    def posterior_period(self, burn: int = 0, thin: int = 1) -> float:
        df = self.posterior_samples(burn, thin, derived_parameters=False)
        tccols = [c for c in df.columns if 'tc' in c]
        tcs = median(df[tccols], 0)
        return mean((tcs[1:] - tcs[0]) / (self.tnumber[1:] - self.tnumber[0]))

    def plot_ttvs(self, burn=0, thin=1, axs=None, figsize=None, bwidth=0.8, fmt='h', windows=None):
        assert fmt in ('d', 'h', 'min')
        multiplier = {'d': 1, 'h': 24, 'min': 1440}
        ncol = 1 if windows is None else len(windows)
        fig, axs = (None, axs) if axs is not None else subplots(1, ncol, figsize=figsize, sharey=True)
        df = self.posterior_samples(burn, thin)
        tccols = [c for c in df.columns if 'tc' in c]
        tcs = median(df[tccols], 0)
        lineph = poly1d(polyfit(self.tnumber, tcs, 1))
        tc_linear = lineph(self.tnumber)
        p = multiplier[fmt] * percentile(df[tccols] - tc_linear, [50, 16, 84, 0.5, 99.5], 0)
        setp(axs, ylabel='Transit center - linear prediction [{}]'.format(fmt), xlabel='Transit number')
        if windows is None:
            plot_estimates(self.tnumber, p, axs, bwidth)
            if with_seaborn:
                sb.despine(ax=axs, offset=15)
        else:
            setp(axs[1:], ylabel='')
            for ax, w in zip(axs, windows):
                m = (self.tnumber > w[0]) & (self.tnumber < w[1])
                plot_estimates(self.tnumber[m], p[:, m], ax, bwidth)
                setp(ax, xlim=w)
                if with_seaborn:
                    sb.despine(ax=ax, offset=15)
        if fig:
            fig.tight_layout()
        return axs



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Optional, Union
from pathlib import Path

from numba import njit
from numpy import zeros, interp, pi, ndarray
from numpy.random import randint

from .ldmodel import LDModel

try:
    from ldtk import LDPSetCreator
except ImportError:
    class LDPSetCreator:
        def __init__(self, *nargs, **kwargs):
            raise ModuleNotFoundError('LDTkLDModel requires LDTk.')

@njit
def ntrapz(x, y):
    npt = x.size
    ii = 0.0
    for i in range(1, npt):
        ii += (x[i] - x[i - 1]) * 0.5 * (y[i] + y[i - 1])
    return ii


@njit
def eval_ldm(emu, mu, z, ldps, npv, nsamples):
    npb = ldps.shape[0]
    ldp = zeros((npv, npb, emu.size))
    ldi = zeros((npv, npb))
    iis = randint(0, nsamples, size=npv)

    for ipv in range(npv):
        for ipb in range(npb):
            ldp[ipv, ipb] = interp(emu, mu, ldps[ipb, iis[ipv]])
            ldi[ipv, ipb] = -2 * pi * ntrapz(z, z * ldps[ipb, iis[ipv]])
    return ldp, ldi


@njit
def eval_ldm_frozen(emu, mu, z, mldps, npv):
    npb = mldps.shape[0]
    ldp = zeros((npv, npb, emu.size))
    ldi = zeros((npv, npb))

    for ipb in range(npb):
        ldp[0, ipb] = interp(emu, mu, mldps[ipb])
        ldi[0, ipb] = -2 * pi * ntrapz(z, z * mldps[ipb])

    for i in range(1, npv):
        ldp[i] = ldp[0]
        ldi[i] = ldi[0]

    return ldp, ldi


class LDTkLDModel(LDModel):
    def __init__(self, pbs: Tuple, teff: Tuple[float, float], logg: Tuple[float, float], z: Tuple[float, float],
                 samples: int = 500, frozen: bool = False, cache: Optional[Union[str, Path]] = None,
                 dataset: str = 'vis-lowres'):
        super().__init__()
        self._sc = LDPSetCreator(teff, logg, z, pbs, cache=cache, dataset=dataset)
        self._ps = self._sc.create_profiles(samples)
        self._i = 0

        self.npb = len(pbs)
        self.nsamples = samples
        self.frozen = frozen
        self.z = self._ps._z
        self.mu = self._ps._mu
        self.profiles = self._ps._ldps
        self.mean_profiles = self._ps.profile_averages

    def thaw(self):
        self.frozen = False

    def freeze(self):
        self.frozen = True

    def __call__(self, mu: ndarray, x: Optional[ndarray] = None) -> Tuple[ndarray, ndarray]:
        npv = 1 if x is None else x.shape[0]
        self._i = i = randint(0, self.nsamples)
        if self.frozen:
            return eval_ldm_frozen(mu, self.mu, self.z, self.mean_profiles, npv)
        else:
            return eval_ldm(mu, self.mu, self.z, self.profiles, npv, self.nsamples)

    def _evaluate(self, mu: ndarray, x:ndarray) -> ndarray:
        raise NotImplementedError

    def _integrate(self, x: ndarray) -> float:
        raise NotImplementedError

class LDTkLD(LDTkLDModel):
    ...


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import repeat, atleast_2d, sqrt, inf

from ..orbits import as_from_rhop, i_from_ba
from ..param import PParameter, UniformPrior as UP
from .lpf import BaseLPF, map_ldc


class MCLPF(BaseLPF):

    def _init_p_planet(self):
        """Set up the planet parametrization

        Set up the radius ratio parametrization, one radius ratio per passband.
        """
        pk2 = [PParameter(f'k2_{pb}', f'{pb} area ratio', 'A_s', UP(0.07 ** 2, 0.12 ** 2), (0, inf)) for pb in
               self.passbands]
        self.ps.add_passband_block('k2', 1, self.npb, pk2)
        self._pid_k2 = repeat(self.ps.blocks[-1].start, self.npb)
        self._start_k2 = self.ps.blocks[-1].start
        self._sl_k2 = self.ps.blocks[-1].slice

    def transit_model(self, pv, copy=True):
        """Calculate the transit model

        We need to override the default `BaseLPF.transit_model` method since it assumes passband-independent
        (achromatic) radius ratios. We have one free area ratio parameter per passband, and the mapping
        from the parameter array to transit model parameters needs to be adjusted.
        """
        pv = atleast_2d(pv)
        ldc = map_ldc(pv[:, self._sl_ld])
        radius_ratios = sqrt(pv[:, 4:4 + self.npb])
        zero_epoch = pv[:, 0] - self._tref
        period = pv[:, 1]
        smaxis = as_from_rhop(pv[:, 2], period)
        inclination = i_from_ba(pv[:, 3], smaxis)
        return self.tm.evaluate(radius_ratios, ldc, zero_epoch, period, smaxis, inclination)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.constants import c, k, h
from scipy.optimize import fmin

def fold(time, period, origo=0.0, shift=0.0, normalize=True, clip_range=None):
    """Folds the given data over a given period.

    Parameters
    ----------
    
      time        
      period      
      origo       
      shift       
      normalize   
      clip_range  

    Returns
    -------

      phase       
    """
    tf = ((time - origo)/period + shift) % 1.

    if not normalize:
        tf *= period
        
    if clip_range is not None:
        mask = np.logical_and(clip_range[0]<tf, tf<clip_range[1])
        tf = tf[mask], mask
    return tf


def planck(T, wl):
    """Radiance as a function or black-body temperature and wavelength.

    Parameters
    ----------

      T   : Temperature  [K]
      wl  : Wavelength   [m]

    Returns
    -------

      B   : Radiance
    """
    return 2*h*c**2/wl**5 / (np.exp(h*c/(wl*k*T))-1)


def contamination_bb(c1, T, wl1, wl2):
    """Contamination from a third object radiating as a black-body given a contamination estimate in a reference wavelength. 

    Parameters
    ----------

      c1   : Contamination in the reference wavelength [-]
      T    : Temperature                               [K]
      wl1  : Reference wavelength                      [m]
      wl2  : Target wavelength                         [m]

    Returns
    -------

      c2  : Contamination in the given wavelength      [-]
    """
    B1   = planck(T, wl1)
    B2   = planck(T, wl2)

    return  c1*(B2/B1)


def nonlinear_ld_to_general_ld(ldc):
    def nl(mu,ld):
        return 1. - np.sum([ld[i-1]*(1.-mu**(0.5*i)) for i in range(1,5)], axis=0)
    def gn(mu,ld):
        return 1. - np.sum([ld[i-1]*(1.-mu**i) for i in range(1,5)], axis=0)

    mu = np.linspace(0,1,200)

    pvg = fmin(lambda pv:sum((Inl-gn(mu,pv))**2), ld[::2])



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional
from numba import prange, njit
from numpy import atleast_2d, zeros, dot, array, concatenate, squeeze, r_, inf, unique, full, ndarray, ones

from ...param import LParameter, NormalPrior as NP

@njit(parallel=True)
def linear_model(pvp, cids, cstart, ncov, cova):
    pvp = atleast_2d(pvp)
    npv = pvp.shape[0]
    npt = cids.size
    bl = zeros((npv, npt))
    for ipv in prange(npv):
        ii = 0
        for ipt in range(npt):
            ilc = cids[ipt]  # Current local light curve index
            cst = cstart[ilc]  # Start of the coefficients in the parameter vector
            nc = ncov[ilc]  # Number of covariates for the current light curve
            bl[ipv, ipt] = pvp[ipv, cst] + dot(pvp[ipv, cst + 1:cst + nc + 1], cova[ii:ii + nc])
            ii += nc
    return bl


class LinearModelBaseline:
    def __init__(self, lpf, name: str = 'lm', lcids=None):
        self.name = name
        self.lpf = lpf

        if lpf.lcids is None:
            raise ValueError('The LPF data needs to be initialised before initialising LinearModelBaseline.')

        self.init_data(lcids)
        self.init_parameters()

    def init_data(self, lcids = None):
        self.lcids = lcids if lcids is not None else unique(self.lpf.lcids)
        self.nlc = self.lcids.size
        self.ncov = array(
            [self.lpf.covariates[lcid].shape[1] for lcid in self.lcids])  # Number of covariates per light curve
        self.cova = concatenate([self.lpf.covariates[lcid].ravel() for lcid in self.lcids])  # Flattened covariate vector
        self.cids = concatenate([full((self.lpf.lcids == lcid).sum(), i) for i, lcid in enumerate(self.lcids)])
        self._bl_coeff_start = r_[[0], self.ncov + 1].cumsum()  # Parameter vector start index for the coefficients

        self.mask = zeros(self.lpf.lcids.size, bool)
        for lcid in self.lcids:
            self.mask[self.lpf.lcslices[lcid]] = 1

        if hasattr(self.lpf, 'ins'):
            self.ins = [self.lpf.ins[lcid] for lcid in self.lcids]
        else:
            self.ins = self.nlc * ['']

        if hasattr(self.lpf, 'piis'):
            self.piis = [self.lpf.piis[lcid] for lcid in self.lcids]
        else:
            self.piis = zeros(self.nlc, int)

    def init_parameters(self):
        """Baseline parameter initialisation.
        """
        bls = []
        for i, tn in enumerate(range(self.nlc)):
            ins = self.ins[i] + '_'
            pii = self.piis[i]

            fstd = self.lpf.fluxes[i].std()
            bls.append(LParameter(f'{self.name}_i_{ins}{pii}', f'{self.name}_intercept_{ins}{pii}', '', NP(1.0, fstd), bounds=(-inf, inf)))
            for ipoly in range(1, self.ncov[i] + 1):
                bls.append(
                    LParameter(f'{self.name}_s_{ins}{pii}_{ipoly}', f'{self.name}_c_{ins}{pii}_{ipoly}', '', NP(0.0, fstd), bounds=(-inf, inf)))
        self.lpf.ps.thaw()
        self.lpf.ps.add_global_block(self.name, bls)
        self.lpf.ps.freeze()
        self.pv_slice = self.lpf.ps.blocks[-1].slice
        self.pv_start = self.lpf.ps.blocks[-1].start
        setattr(self.lpf, f"_sl_{self.name}", self.pv_slice)
        setattr(self.lpf, f"_start_{self.name}", self.pv_start)

    def __call__(self, pvp, bl: Optional[ndarray] = None):
        pvp = atleast_2d(pvp)
        if bl is None:
            bl = ones((pvp.shape[0], self.lpf.timea.size))
        else:
            bl = atleast_2d(bl)
        bl[:, self.mask] += linear_model(pvp[:, self.pv_slice], self.cids, self._bl_coeff_start, self.ncov, self.cova) - 1.
        return squeeze(bl)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2020  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit
from numpy import ones, pi, sqrt, log, exp, zeros, power, log2


@njit(fastmath=True)
def ld_uniform(mu, pv):
    return ones(mu.size)


@njit(fastmath=True)
def ldi_uniform(pv):
    return pi


@njit(fastmath=True)
def ld_linear(mu, pv):
    return 1. - pv[0] * (1. - mu)


@njit(fastmath=True)
def ldi_linear(pv):
    return 2 * pi * 1 / 6 * (3 - 2 * pv[0])


@njit(fastmath=True)
def ld_quadratic(mu, pv):
    return 1. - pv[0] * (1. - mu) - pv[1] * (1. - mu) ** 2


@njit(fastmath=True)
def ldi_quadratic(pv):
    return 2 * pi * 1 / 12 * (-2 * pv[0] - pv[1] + 6)


@njit(fastmath=True)
def ld_quadratic_tri(mu, pv):
    a, b = sqrt(pv[0]), 2 * pv[1]
    u, v = a * b, a * (1. - b)
    return 1. - u * (1. - mu) - v * (1. - mu) ** 2


@njit(fastmath=True)
def ldi_quadratic_tri(pv):
    a, b = sqrt(pv[0]), 2 * pv[1]
    u, v = a * b, a * (1. - b)
    return 2 * pi * 1 / 12 * (-2 * u - v + 6)


@njit(fastmath=True)
def ld_nonlinear(mu, pv):
    return 1. - pv[0] * (1. - sqrt(mu)) - pv[1] * (1. - mu) - pv[2] * (1. - power(mu, 1.5)) - pv[3] * (1. - mu ** 2)


@njit(fastmath=True)
def ld_general(mu, pv):
    ldp = zeros(mu.size)
    for i in range(pv.size):
        ldp += pv[i] * (1.0 - mu ** (i + 1))
    return ldp


@njit(fastmath=True)
def ld_square_root(mu, pv):
    return 1. - pv[0] * (1. - mu) - pv[1] * (1. - sqrt(mu))


@njit(fastmath=True)
def ld_logarithmic(mu, pv):
    return 1. - pv[0] * (1. - mu) - pv[1] * mu * log(mu)


@njit(fastmath=True)
def ld_exponential(mu, pv):
    return 1. - pv[0] * (1. - mu) - pv[1] / (1. - exp(mu))


@njit(fastmath=True)
def ld_power_2(mu, pv):
    return 1. - pv[0] * (1. - mu ** pv[1])

@njit(fastmath=True)
def ld_power_2_pm(mu, pv):
    c = 1 - pv[0] + pv[1]
    a = log2(c/pv[1])
    return 1. - c * (1. - mu**a)

@njit
def evaluate_ld(ldm, mu, pvo):
    if pvo.ndim == 1:
        pv = pvo.reshape((1, 1, -1))
    elif pvo.ndim == 2:
        pv = pvo.reshape((1, pvo.shape[1], -1))
    else:
        pv = pvo

    npv = pv.shape[0]
    npb = pv.shape[1]
    ldp = zeros((npv, npb, mu.size))
    for ipv in range(npv):
        for ipb in range(npb):
            ldp[ipv, ipb, :] = ldm(mu, pv[ipv, ipb])
    return ldp


@njit
def evaluate_ldi(ldi, pvo):
    if pvo.ndim == 1:
        pv = pvo.reshape((1, 1, -1))
    elif pvo.ndim == 2:
        pv = pvo.reshape((1, pvo.shape[1], -1))
    else:
        pv = pvo

    npv = pv.shape[0]
    npb = pv.shape[1]
    istar = zeros((npv, npb))
    for ipv in range(npv):
        for ipb in range(npb):
            istar[ipv, ipb] = ldi(pv[ipv, ipb])
    return istar


#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2021  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from os.path import join
import astropy.io.fits as pf
import pandas as pd
from numba import njit
from numpy import linspace, zeros, array, arange, exp
from pkg_resources import resource_filename
from scipy.interpolate import RegularGridInterpolator
from tqdm.auto import tqdm

__all__ = ['husser2013_file', 'compute_averaged_husser2013_table', 'create_husser2013_interpolator', 'read_husser2013_table']

husser2013_file = resource_filename(__name__, join("data", "avg_husser2013.fits"))


def get_teff(f):
    return int(f.name[3:8])


def gather_files(datadir):
    files = {}
    for f in sorted(datadir.glob('lte*HiRes.fits')):
        teff = get_teff(f)
        if teff not in files:
            files[teff] = []
        files[teff].append(f)
    return files


def read_spectrum(fname, wl = None):
    fl = d = pf.getdata(fname).astype('d')
    if wl is None:
        h = pf.getheader(fname)
        wl = 0.1*exp(h['CRVAL1'] + arange(h['NAXIS1'])*h['CDELT1'])
    return wl, fl


@njit
def bin_spectrum(wl, fl, lmin: int = 300, lmax: int = 2500):
    bwl = linspace(lmin, lmax, lmax - lmin + 1)
    bfl = zeros(bwl.size)
    istart, iend = 0, 0
    for ibin in range(bfl.size):
        cwl = bwl[ibin]
        n = 0
        while wl[istart] < cwl - 0.5:
            istart += 1
        iend = istart
        while wl[iend] < cwl + 0.5 and iend < wl.size - 1:
            bfl[ibin] += fl[istart]
            n += 1
            iend += 1
        istart = iend
        if n > 0:
            bfl[ibin] /= n
    return bwl, bfl


def average_over_teff(files, teff):
    wl, fl = None, None
    for f in tqdm(files[teff], leave=False):
        if wl is None:
            wl, fl = read_spectrum(f)
        else:
            _, flc = read_spectrum(f, wl)
            fl += flc
    fl /= len(files[teff])
    return wl, fl


def create_table(files):
    teffs = list(files.keys())
    spectra = []
    for teff in tqdm(teffs):
        wl, fl = average_over_teff(files, teff)
        bwl, bfl = bin_spectrum(wl, fl)
        spectra.append(bfl)
    spectra = array(spectra)
    return pd.DataFrame(spectra, index=pd.Index(teffs, name='TEff [K]'),
                        columns=pd.Index(bwl, name='Wavelength [nm]'))


def write_table(df):
    hdu0 = pf.PrimaryHDU(data=df.values)
    hdu0.header['CRVAL1'] = (300, 'Wavelength of the reference pixel')
    hdu0.header['CDELT1'] = (1, 'Wavelength grid step size')
    hdu0.header['CRPIX1'] = (1, 'Reference pixel coordinates')
    hdu0.header['CTYPE1'] = ('WAVE', 'Type of wavelength grid')
    hdu0.header['CUNIT1'] = ('nm', 'Unit of wavelength grid')
    hdu0.header['CTYPE2'] = ('TEFF', 'TEff array defind in "TEFF" HDU')
    cols = pf.ColDefs([pf.Column(name='TEff', format='E', unit='K', array=df.index.values)])
    hdu1 = pf.BinTableHDU.from_columns(cols)
    hdu1.header['EXTNAME'] = 'TEFF'
    pf.HDUList([hdu0, hdu1]).writeto(husser2013_file, overwrite=True)


def read_husser2013_table():
    spectra = pf.getdata(husser2013_file)
    teff = pf.getdata(husser2013_file, 1)['TEff'].astype('d')
    wl = pf.getval(husser2013_file, 'CRVAL1') + arange(spectra.shape[1]) * pf.getval(husser2013_file, 'CDELT1')
    return pd.DataFrame(spectra, index=pd.Index(teff, name='TEff [K]'), columns=pd.Index(wl, name='Wavelength [nm]'))


def create_husser2013_interpolator():
    df = read_husser2013_table()
    rgi = RegularGridInterpolator((df.index.values, df.columns.values.astype('d')), df.values)
    return rgi


def compute_averaged_husser2013_table(datadir):
    files = gather_files(datadir)
    df = create_table(files)
    write_table(df)



#  PyTransit: fast and easy exoplanet transit modelling in Python.
#  Copyright (C) 2010-2019  Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import asarray, zeros
import math as mt

class LDParameterization(object):
    __slots__ = ('coefs',)
    name  = ''
    ncoef = 0

    def __init__(self, coefs=None):
        self.coefs = zeros(self.ncoef) if coefs is None else asarray(coefs)
        assert self.coefs.size == self.ncoef
    
    def __str__(self):
        return '{} : {}'.format(self.name, self.coefs)
    
    def __repr__(self):
        return '{} : {}'.format(self.name, self.coefs)

    def map_from(self, other):
        raise NotImplementedError

    @property
    def uniform(self):
        raise NotImplementedError
    
    @property
    def linear(self):
        raise NotImplementedError
        
    @property
    def quadratic(self):
        raise NotImplementedError
        
    @property    
    def triangular(self):
        raise NotImplementedError
        
       
class UniformLD(LDParameterization):
    """Uniform limb darkening parameterization
    """
    __slots__ = ('coefs',)
    name  = 'uniform'
    ncoef = 0
    
    @property
    def uniform(self):
        return self
    
    @property    
    def linear(self):
        return LinearLD([0.])
    
    @property    
    def quadratic(self):
        return QuadraticLD([0., 0.])
    
    @property
    def triangular(self):
        return TriangularQLD([0., 0.])
    
    
class LinearLD(LDParameterization):
    """Linear limb darkening parameterization
    """
    __slots__ = ('coefs')
    name = 'linear'
    ncoef = 1

    def map_from(self, other):
        self.coefs[:] = other.linear.coefs
    
    @property
    def linear(self):
        return self
    
    @property    
    def quadratic(self):
        return QuadraticLD([self.coefs[0], 0])
    
    @property
    def triangular(self):
        return self.quadratic.triangular
    
    
class QuadraticLD(LDParameterization):
    """Quadratic limb darkening parameterization
    """
    __slots__ = ('coefs',)
    name = 'quadratic'
    ncoef = 2

    def map_from(self, other):
        self.coefs[:] = other.quadratic.coefs
    
    @property    
    def quadratic(self):
        return self
    
    @property      
    def triangular(self):
        u,v = self.coefs
        return TriangularQLD([(u+v)**2, u/(2*(u+v))])
        
        
class TriangularQLD(QuadraticLD):
    """Triangular quadratic parametrization by Kipping (2013)
    
    Kipping, D. MNRAS, 435 (3) pp. 2152-2160, 2013
    """
    __slots__ = ('coefs',)
    name = 'triangular quadratic'

    def map_from(self, other):
        self.coefs[:] = other.triangular.coefs
    
    @property
    def quadratic(self):    
        a,b = mt.sqrt(self.coefs[0]), 2*self.coefs[1]
        return QuadraticLD([a*b, a*(1-b)])
    
    @property
    def triangular(self):
        return self
